# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_kermac import Stack, PtxInstruction
from mm_kermac.hyper_semiring import HyperSemiringKernel

import torch

class NormL2:
    def __init__(self):
        self.hyper_semiring = \
            HyperSemiringKernel(
                multiply_lambda=\
                    lambda reg_a, reg_b, _: [
                        reg_a,
                        reg_b,
                        PtxInstruction.sub_ftz_f32,           # diff = reg_b - reg_a [diff]
                        Stack.f32.dup,                        # [diff, diff]
                        PtxInstruction.mul_ftz_f32            # [diff * diff]
                    ],
                accumulate_lambda=\
                    lambda reg_diff, reg_c, _: [
                        reg_diff,
                        reg_c,
                        PtxInstruction.add_ftz_f32
                    ],
                epilogue_lambda=\
                    lambda reg_e, _: [
                        reg_e,
                        PtxInstruction.sqrt_approx_ftz_f32
                    ]
            )
        
    def __call__(
        self,
        x : torch.Tensor,
        z : torch.Tensor,
        out : torch.Tensor = None,
        try_to_align : bool = False,
        debug = False
    ):
        return self.hyper_semiring(
            a = x,
            b = z,
            out = out,
            hyper_dict=None,
            try_to_align = try_to_align,
            debug = debug
        )



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_ptx.ptx_inject import DataTypeInfoEnum

from enum import auto, unique

# Define the types the PTX Inject system might find inside the cuda PTX_INJECT annotations
# name: the name in the PTX_INJECT cuda annotations.
# ptx_mov_name: the type PTX Inject for the mov instructions.
# ptx_register_char: The inline ptx constraint letter for the PTX register type. (see inline ptx document)
# register_cast_str: The cast string for the register in the inline PTX. For example, 'half'
# is a struct that needs to be cast.
#   half x;
#   asm(
#    ".."
#    : "=h"(*(unsigned short*)&x)
#   )
# Being able to define these types dynamically in python allows customization on the user side.
# For example, you could define a type
# B1X32 = (auto(), "b1x32", "b32", "b32", 'r')
# That specifies to you that this value contains 32 individual bits per register.
# You can then add this as a type in your CUDA code PTX Inject annotations.
# When traversing the PTX Inject structure you could prefer to dynamically
# generate PTX only relevant to this being a B1X32 data type.
# i.e. PTX "and.b32", "or.b32", "shl.b32", "shr.b32", etc...
@unique
class DataTypeInfo(DataTypeInfoEnum):
#   enum_name,  enum_idx,   name,       register_type,  ptx_mov_name,   ptx_register_char,  register_cast_str
#   F16   =     (auto(),    "f16",      "f16",          "b16",          'h',                "*(unsigned short*)&"   )
#   F16X2 =     (auto(),    "f16x2",    "f16x2",        "b32",          'r',                "*(unsigned int*)&"     )
    S32   =     (auto(),    "s32",      "s32",          "s32",          'r')
    U32   =     (auto(),    "u32",      "u32",          "u32",          'r')
    F32   =     (auto(),    "f32",      "f32",          "f32",          'f')
    B32   =     (auto(),    "b32",      "b32",          "b32",          'r')



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import argparse
import mm_kermac
import torch

from mm_kermac.helpers import (
    compare_tensors,
    torch_einsum_L2_norm,
    torch_inefficient_p_norm
)
from mm_kermac.stack_ptx_types import Stack, PtxInstruction

import mm_kermac.hyper_semiring_gradient as hyper_semiring_gradient

def parse_args():
    """Parse command-line arguments for matrix dimensions, p-norm, and flags."""
    parser = argparse.ArgumentParser(description="Run kermac.cdist_t with configurable parameters")
    parser.add_argument('-m','--M', type=int, default=20, help='Number of rows in data_z (default: 20)')
    parser.add_argument('-n','--N', type=int, default=16, help='Number of columns in data (default: 16)')
    parser.add_argument('-o','--O', type=int, default=3, help='Number of channels in coefficents (default: 3)')
    parser.add_argument('-k','--K', type=int, default=100, help='Number of rows in data_x (default: 100)')
    parser.add_argument('-l','--L', type=int, default=2, help='Number of batches in each dimension (default: 2)')
    parser.add_argument('-d','--debug', default=False, action='store_true', help='Enable debug output (default: True)')
    return parser.parse_args()

def main():
    args = parse_args()
    M, N, O, K, L = args.M, args.N, args.O, args.K, args.L
    debug = args.debug

    device = torch.device('cuda')
    timer = mm_kermac.CudaTimer()

    size_M = M
    size_D = N
    size_C = O
    size_N = K
    size_L = L

    tensor_A = torch.randn(size_L,size_N,size_M,device=device) # M-major # M-major 
    tensor_B = torch.randn(size_L,size_D,size_N,device=device) # N-major # K-major
    tensor_C = torch.randn(size_L,size_C,size_N,device=device) # N-major # K-major
    tensor_D = torch.randn(size_L,size_D,size_M,device=device) # M-major # M-Major

    coefs =         tensor_C
    grad_kernel_matrix = tensor_A
    x =             tensor_B
    z =             tensor_D

    torch.cuda.synchronize()

    epsilon = 1e-3

    print('Running Norm L1 Grad')
    norm_L1 = hyper_semiring_gradient.NormL1()
    kermac_out = \
        norm_L1(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            debug = debug
        )
    
    torch_out = \
        torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=1.0,
            eps=0.0,
        )
    compare_tensors(kermac_out, torch_out)

    print('Running Norm L2 Grad')
    norm_L2 = hyper_semiring_gradient.NormL2()
    kermac_out = \
        norm_L2(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            debug = debug
        )
    
    torch_out = \
        torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=2.0,
            eps=0.0,
        )
    einsum_out = \
        torch_einsum_L2_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix
        )
    print("   Against Inefficient P-norm")
    compare_tensors(kermac_out, torch_out)
    print("   Against Torch Einsum L2-norm")
    compare_tensors(kermac_out, einsum_out)

    p_power_grad = torch.tensor([1.3 - 1.0], dtype=torch.float32, device=device)

    print('Running Norm P=1.3 Grad')
    norm_Lp = \
        hyper_semiring_gradient.NormLp(
            epsilon=epsilon,
        )
    kermac_out = \
        norm_Lp(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p_power_grad=p_power_grad,
            debug = debug
        )
    
    torch_out = \
        torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=1.3,
            eps=epsilon,
        )
    compare_tensors(kermac_out, torch_out)

if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import json
import base64

class FunctionDBKey:
    """Represents the key structure for the function database."""
    def __init__(
        self,
        package_name: str,
        package_version: str,
        cuda_version: str,
        arch: str,
        function_name: str,
        cuda_source_hash: bytes
    ):
        self.package_name = package_name
        self.package_version = package_version
        self.cuda_version = cuda_version
        self.arch = arch
        self.function_name = function_name
        self.cuda_source_hash = cuda_source_hash

    def to_bytes(self) -> bytes:
        """Serialize the key to a bytes object for LMDB storage."""
        key_dict = {
            'package_name': self.package_name,
            'package_version': self.package_version,
            'cuda_version': self.cuda_version,
            'arch': self.arch,
            'function_name': self.function_name,
            'cuda_source_hash': self.cuda_source_hash
        }
        return json.dumps(key_dict, sort_keys=True).encode('utf-8')

    @classmethod
    def from_bytes(cls, data: bytes) -> 'FunctionDBKey':
        """Deserialize bytes to a FunctionDBKey object."""
        key_dict = json.loads(data.decode('utf-8'))
        return cls(
            package_name=key_dict['package_name'],
            package_version=key_dict['package_version'],
            cuda_version=key_dict['cuda_version'],
            arch=key_dict['arch'],
            function_name=key_dict['function_name'],
            cuda_source_hash=key_dict['cuda_source_hash']
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FunctionDBKey):
            return False
        return (
            self.package_name == other.package_name and
            self.package_version == other.package_version and
            self.cuda_version == other.cuda_version and
            self.arch == other.arch and
            self.function_name == other.function_name and
            self.cuda_source_hash == other.cuda_source_hash
        )

    def __repr__(self) -> str:
        return (f"FunctionDBKey(package_name={self.package_name}, "
                f"package_version={self.package_version}, "
                f"cuda_version={self.cuda_version}, "
                f"arch={self.arch}, "
                f"function_name={self.function_name}, "
                f"cuda_source_hash={self.cuda_source_hash})")


# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import torch

def compare_tensors(
    tensor_a,
    tensor_b
):
    diff = tensor_a - tensor_b
    squared_diff = diff ** 2
    mse = torch.mean(squared_diff)
    rmse = torch.sqrt(mse).item()

    abs_error = torch.abs(diff)
    max_abs_error = torch.max(abs_error).item()
    mean_abs_error = torch.mean(abs_error).item()

    print(f"\tRoot Mean Squared Error:     {rmse:.6e}")
    print(f"\tMax Absolute Error:          {max_abs_error:.6e}")
    print(f"\tMean Absolute Error:         {mean_abs_error:.6e}")

def torch_einsum_L2_norm(
    x: torch.Tensor,
    z: torch.Tensor,
    coefs: torch.Tensor,
    grad_kernel_matrix: torch.Tensor
):
    torch_out = torch.einsum('bli,bij,bjd->bljd', coefs, grad_kernel_matrix, z.permute(0,2,1)) - torch.einsum('bli,bij,bid->bljd', coefs, grad_kernel_matrix, x.permute(0,2,1))
    return torch_out.permute(0,1,3,2)

def torch_inefficient_p_norm(
    x: torch.Tensor,                    # (B, D, I)   == tensor_mini_B
    z: torch.Tensor,                    # (B, D, J)   == tensor_mini_D
    coefs: torch.Tensor,                # (B, L, I)   == tensor_mini_C
    grad_kernel_matrix: torch.Tensor,   # (B, I, J)   == tensor_mini_A
    p: float,
    eps: float = 0.0  
):
    """
    Computes the batched version of:
        out[o, n, m] = sum_k c[o,k] * a[k,m] * sign(d[n,m] - b[n,k]) * |d[n,m] - b[n,k]|^(p-1)

    Batched shapes:
      a: (B, I, J)   -> K=I, M=J
      b: (B, D, I)   -> N=D, K=I
      c: (B, L, I)   -> O=L, K=I
      d: (B, D, J)   -> N=D, M=J
      returns: (B, L, D, J)
    """
    # diff[b, n, m, k] = d[b, n, m] - b[b, n, k]
    diff = z[:, :, :, None] - x[:, :, None, :]      # (B, D, J, I)

    if p == 2.0 and eps == 0.0:
        # sign(x)*|x|^(p-1) == x when p=2
        g = diff                                     # (B, D, J, I)
    else:
        g = torch.sign(diff) * diff.abs().clamp_min(eps).pow(p - 1.0)  # (B, D, J, I)

    # out[b, o, n, m] = sum_k c[b,o,k] * a[b,k,m] * g[b,n,m,k]
    # c: (B, L, I), a: (B, I, J), g: (B, D, J, I)  -> einsum over k=I
    out = torch.einsum('bli,bij,bdji->bldj', coefs, grad_kernel_matrix, g)  # (B, L, D, J)
    return out



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_kermac import Stack, PtxInstruction
from mm_kermac.hyper_semiring_gradient import HyperSemiringGradientKernel

import torch

class NormL1:
    def __init__(self):
        self.hyper_semiring_gradient = \
            HyperSemiringGradientKernel(
                multiply_lambda=\
                    lambda reg_d, reg_b, reg_a, _: [
                        reg_b,
                        reg_d,
                        PtxInstruction.sub_ftz_f32,
                        Stack.f32.constant(1.0),
                        Stack.f32.swap,
                        PtxInstruction.copysign_f32,
                        reg_a,
                        PtxInstruction.mul_ftz_f32
                    ],
                accumulate_lambda=\
                    lambda reg_c, reg_diff, reg_e, _: [
                        reg_c,
                        reg_diff,
                        PtxInstruction.mul_ftz_f32,
                        reg_e,
                        PtxInstruction.add_ftz_f32
                    ],
                epilogue_lambda=\
                    lambda reg_e, _: [
                        reg_e
                    ]
            )
        
    def __call__(
        self,
        x : torch.Tensor,
        z : torch.Tensor,
        coefs : torch.Tensor,
        grad_kernel_matrix : torch.Tensor,
        out : torch.Tensor = None,
        debug = False
    ):
        return self.hyper_semiring_gradient(
            a = grad_kernel_matrix,
            b = x,
            c = coefs,
            d = z,
            hyper_dict=None,
            out = out,
            debug = debug
        )



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from .hyper_semiring_gradient import (
    HyperSemiringGradientKernel
)

from .zoo.norm_L1   import NormL1
from .zoo.norm_L2   import NormL2
from .zoo.norm_Lp   import NormLp


# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from cuda.core.experimental import Device, LaunchConfig, launch, ObjectCode

from mm_kermac.ptx_inject_cache.ptx_inject_cache import *
from mm_kermac.ptx_inject_cache.ptx_inject_types import DataTypeInfo
from mm_kermac.common import *

import torch
import numpy as np

import mm_ptx.ptx_inject as ptx_inject
import mm_ptx.stack_ptx as stack_ptx

from mm_kermac.stack_ptx_types import Stack
from mm_kermac.stack_ptx_types import compiler as stack_ptx_compiler

class HyperSemiringKernel():
    def __init__(
        self,
        multiply_lambda,
        accumulate_lambda,
        epilogue_lambda,
        execution_limit=100,
        max_ast_size=100,
        max_ast_to_visit_stack_depth=20,
        stack_size=128,
        max_frame_depth=4
    ):
        self.multiply_lambda = multiply_lambda
        self.accumulate_lambda = accumulate_lambda
        self.epilogue_lambda = epilogue_lambda

        self.cubin_kernels_dict = {}

        self.execution_limit = execution_limit
        self.max_ast_size = max_ast_size
        self.max_ast_to_visit_stack_depth = max_ast_to_visit_stack_depth
        self.stack_size = stack_size
        self.max_frame_depth = max_frame_depth

    def _validate_and_collect_hypers(
        self,
        hyper_dict: dict[str, torch.Tensor] | None,
        L: int,
        tensor_device: torch.device
    ):
        """
        - Merges batch size over all hyper tensors (expected scalar or length-L).
        - Ensures each value is a CUDA float32 tensor on the same device.
        - Returns updated L and a list[(name, tensor)] preserving dict insertion order.
        """
        if hyper_dict is None:
            return L, []

        if not isinstance(hyper_dict, dict):
            raise TypeError("hyper_dict must be a dict[str, torch.Tensor]")

        # 1) Merge batch size
        for name, t in hyper_dict.items():
            if not isinstance(t, torch.Tensor):
                raise TypeError(f"hyper_dict['{name}'] must be a torch.Tensor")
            # Scalars (0-D) or per-batch (1-D) supported, same behavior as inner_p/outer_p/etc
            L = merge_batch_size(f"hyper_dict['{name}']", L, t, expected_dims=0, can_be_none=False)

        # 2) Dtype / CUDA / device checks
        for name, t in hyper_dict.items():
            if t.dtype != torch.float32:
                raise TypeError(f"hyper_dict['{name}'] must have dtype torch.float32")
            if not t.is_cuda:
                raise ValueError(f"hyper_dict['{name}'] must be on a CUDA device")
            if t.device != tensor_device:
                raise ValueError(
                    f"hyper_dict['{name}'] must be on device {tensor_device}, got {t.device}"
                )

        # Keep the (name, tensor) pairs in insertion order
        return L, list(hyper_dict.items())

    def _render_function_name(
        self,
        num : int,
        majorness_A,
        majorness_B,
        align_A,
        align_B,
    ):
        kernel_name_str = f'cute_hyper_semiring_{num}'
        template_parameters = [
            f'Majorness::{majorness_A.name}',
            f'Majorness::{majorness_B.name}',
            f'Alignment::{align_A.name}',
            f'Alignment::{align_B.name}'
        ]
        function_name = f'{kernel_name_str}<{",".join(template_parameters)}>'
        return function_name
    
    def _create_hyper_semiring_template(
        self,
        num : int
    ):
        template_source_file = get_local_cuda_kernel_dir() / 'template_hyper_semiring.cuh'

        with open(template_source_file, "r") as f:
            content = f.read()
            num_sites = 11
            
            templates = [
                "\n\t".join(f"class HYPER{i}Stride," for i in range(num)),
                f"{num}",
                "\n\t".join(f", T* HYPER{i}, HYPER{i}Stride dHYPER{i}" for i in range(num)),
                "\n\t".join(f" Tensor mHYPER{i} = make_tensor(make_gmem_ptr(HYPER{i}), select<3>(shape_MNKL), dHYPER{i});" for i in range(num)),
                "\n\t".join(f"T hyper{i} = mHYPER{i}(bidz);" for i in range(num)),
                "\n\t\t\t\t\t\t".join(f"in f32 hyper{i}" for i in range(num)),
                "\n\t\t\t\t\t\t".join(f"in f32 hyper{i}" for i in range(num)),
                "\n\t\t\t".join(f"in f32 hyper{i}" for i in range(num)),
                "\n\t".join(f", float *hyper{i},    uint64_t batch_stride_hyper{i}" for i in range(num)),
                "\n\t".join(f"auto d_hyper{i} = make_stride(batch_stride_hyper{i});" for i in range(num)),
                "\n\t\t".join(f", hyper{i}, d_hyper{i}" for i in range(num)),
            ]

            for i in range(num_sites):
                content = content.replace(f"@{i}@", templates[i])

            # # Prints the generated template
            # with open(get_top_level_repo_dir('') / f'rendered_hyper_semiring_{num}.cuh', 'w') as file:
            #     file.write(content)

            return content
    
    def __call__(
        self,
        a : torch.Tensor,
        b : torch.Tensor,
        hyper_dict : dict[str, torch.Tensor] = None,
        out : torch.Tensor = None,
        try_to_align : bool = False,
        debug = False
    ):
        L = 1
        L = merge_batch_size('a', L, a, expected_dims=2, can_be_none=False)
        L = merge_batch_size('b', L, b, expected_dims=2, can_be_none=False)
        L = merge_batch_size('out', L, out, expected_dims=2, can_be_none=True)
            
        # Check if inputs are tensors
        if not isinstance(a, torch.Tensor) or not isinstance(b, torch.Tensor):
            raise TypeError("a and b must be PyTorch tensors")
        if out is not None and not isinstance(out, torch.Tensor):
            raise TypeError("out must be a PyTorch tensor if provided")

        # Check dtype
        if a.dtype != torch.float32 or b.dtype != torch.float32:
            raise TypeError("a and b must have dtype torch.float32")
        if out is not None and out.dtype != torch.float32:
            raise TypeError("out must have dtype torch.float32")

        # Check CUDA device
        if not a.is_cuda or not b.is_cuda:
            raise ValueError("a and b must be on a CUDA device")
        if out is not None and not out.is_cuda:
            raise ValueError("out must be on a CUDA device")
        
        tensor_device = a.device
        if not all(x.device == tensor_device for x in (a, b)):
            raise ValueError(f"All inputs must be on the same CUDA device: got {[x.device for x in (a, b)]}")
        if out is not None and out.device != tensor_device:
            raise ValueError(f"out must be on the same CUDA device as inputs: got {out.device}, expected {tensor_device}")

        # NEW: fold hyper_dict into L and validate each tensor
        L, hyper_items = self._validate_and_collect_hypers(hyper_dict, L, tensor_device)
        num_hypers = len(hyper_items)

        tensor_stats_a = tensor_stats(a)
        tensor_stats_b = tensor_stats(b)

        # Get shapes
        _, M, K_a = tensor_stats_a.shape
        _, N, K_b = tensor_stats_b.shape

        # Check shape consistency
        if K_a != K_b:
            raise ValueError(f"K dimensions must match: got {K_a} for a and {K_b} for b")
        
        K = K_a
        
        if out is not None:
            L_c = 1 if out.dim() == 2 else out.size(0)
            tensor_stats_c = tensor_stats(out)
            _, M_c, N_c = tensor_stats_c.shape
            if (M_c, N_c) != (M,N):
                raise ValueError(f"out must have shape (M={M}, N={N}), got {(M_c, N_c)}")
            if L_c != L and L != 1:
                raise ValueError(f"out must have batch dimension (L={L}), got {(L_c)}")
        else:
            out = torch.zeros((L, M, N), dtype=torch.float32, device=tensor_device)
            tensor_stats_c = tensor_stats(out)
    
        pt_stream = torch.cuda.current_stream()
        pt_device = pt_stream.device
        device = Device(pt_device.index)
        device.set_current()
        stream = PyTorchStreamWrapper(pt_stream)

        if tensor_device != pt_device:
            raise ValueError("cuda stream must be on the same device as the tensors: got {pt_device}, expected {tensor_device}")
        
        if tensor_stats_c.majorness == Majorness.ROW_MAJOR:
            # Swap arguments if output tensor is row major
            # Kernel will dispatch to version with output as col major
            temp_M = M
            M = N
            N = temp_M

            temp_a = a
            a = b
            b = temp_a
            
            temp_tensor_stats_a = tensor_stats_a
            tensor_stats_a = tensor_stats_b
            tensor_stats_b = temp_tensor_stats_a

        align_4_A = Alignment.ALIGN_1 if not try_to_align else tensor_stats_a.alignment
        align_4_B = Alignment.ALIGN_1 if not try_to_align else tensor_stats_b.alignment

        function_name = self._render_function_name(num_hypers,tensor_stats_a.majorness, tensor_stats_b.majorness, align_4_A, align_4_B)
        dict_key = device, function_name
        kernel = None
        if (dict_key in self.cubin_kernels_dict):
            # We found the kernel already loaded for this majorness and alignment
            # (With the same injected PTX)
            kernel = self.cubin_kernels_dict[dict_key]
        else:
            def get_cuda_source_lambda():
                return self._create_hyper_semiring_template(num_hypers)
            # We did not find the kernel for this majorness and alignment
            # and number of hyper parameters requested.
            # Go ahead and grab the PTX out of the database, inject it,
            # get the kernel and cache it locally to the class.
            # If the PTX wasn't already compiled the ptx_inject_cache will 
            # handle it and store in DB.
            ptx_inject_cache = PtxInjectCache(debug)
            annotated_ptx, lowered_name = \
                ptx_inject_cache.get_function(
                    device,
                    function_name, 
                    get_cuda_source_lambda,
                    debug=debug
                )

            inject = ptx_inject.PTXInject(DataTypeInfo, annotated_ptx)

            multiply = inject['multiply']
            accumulate = inject['accumulate']
            epilogue = inject['epilogue']

            assert( multiply['a'].mut_type == ptx_inject.MutType.IN)
            assert( multiply['a'].data_type == DataTypeInfo.F32)

            assert( multiply['b'].mut_type == ptx_inject.MutType.IN)
            assert( multiply['b'].data_type == DataTypeInfo.F32)

            assert( multiply['diff'].mut_type == ptx_inject.MutType.OUT)
            assert( multiply['diff'].data_type == DataTypeInfo.F32)

            assert( accumulate['diff'].mut_type == ptx_inject.MutType.IN)
            assert( accumulate['diff'].data_type == DataTypeInfo.F32)

            assert( accumulate['c'].mut_type == ptx_inject.MutType.MOD)
            assert( accumulate['c'].data_type == DataTypeInfo.F32)

            assert( epilogue['e'].mut_type == ptx_inject.MutType.MOD)
            assert( epilogue['e'].data_type == DataTypeInfo.F32)

            # assert the types for each hyper parameter found in the inject.
            hyper_names = [f"hyper{i}" for i in range(num_hypers)]
            for h in hyper_names:
                for group, gname in (
                    (multiply, "multiply"),
                    (accumulate, "accumulate"),
                    (epilogue, "epilogue")
                ):
                    assert group[h].mut_type == ptx_inject.MutType.IN,  f"{gname}['{h}'] must be IN"
                    assert group[h].data_type == DataTypeInfo.F32,      f"{gname}['{h}'] must be F32"

            registry = stack_ptx.RegisterRegistry()
            # Add the regular register names to the registry.
            registry.add(multiply['a'].reg,         Stack.f32, name='multiply_a')
            registry.add(multiply['b'].reg,         Stack.f32, name='multiply_b')
            registry.add(multiply['diff'].reg,      Stack.f32, name='multiply_diff')
            registry.add(accumulate['diff'].reg,    Stack.f32, name='accumulate_diff')
            registry.add(accumulate['c'].reg,       Stack.f32, name='accumulate_c')
            registry.add(epilogue['e'].reg,         Stack.f32, name='epilogue_e')

            # Add the register names from the hyper parameters.
            for h in hyper_names:
                registry.add(multiply[h].reg,     Stack.f32, name=f"multiply_{h}")
                registry.add(accumulate[h].reg,   Stack.f32, name=f"accumulate_{h}")
                registry.add(epilogue[h].reg,     Stack.f32, name=f"epilogue_{h}")

            registry.freeze()

            multiply_requests =     [registry.multiply_diff]
            accumulate_requests =   [registry.accumulate_c]
            epilogue_requests =     [registry.epilogue_e]

            def _build_stage_register_dict(stage: str) -> dict[str, object]:
                """
                Map user-provided names to the corresponding stage register *invocations*.
                Example: {"my_beta": registry.multiply_hyper0(), "gamma": registry.multiply_hyper1(), ...}
                """
                out: dict[str, object] = {}
                for i, (user_name, _tensor) in enumerate(hyper_items):
                    attr = f"{stage}_hyper{i}"
                    try:
                        sym = getattr(registry, attr)
                    except AttributeError as e:
                        raise AttributeError(f"Registry missing '{attr}' for hyper index {i}") from e
                    out[user_name] = sym  # call to produce the StackPtxInstruction for this input
                return out

            multiply_register_dict   = _build_stage_register_dict("multiply")
            accumulate_register_dict = _build_stage_register_dict("accumulate")
            epilogue_register_dict   = _build_stage_register_dict("epilogue")

            multiply_instructions = \
                self.multiply_lambda(
                    registry.multiply_a,
                    registry.multiply_b,
                    multiply_register_dict
                )
            
            accumulate_instructions = \
                self.accumulate_lambda(
                    registry.accumulate_diff, 
                    registry.accumulate_c,
                    accumulate_register_dict
                )
            
            epilogue_instructions = \
                self.epilogue_lambda(
                    registry.epilogue_e,
                    epilogue_register_dict
                )

            multiply_ptx_stub = \
                stack_ptx_compiler.compile(
                    registry=registry,
                    instructions=multiply_instructions, 
                    requests=multiply_requests,
                    execution_limit=self.execution_limit,
                    max_ast_size=self.max_ast_size,
                    max_ast_to_visit_stack_depth=self.max_ast_to_visit_stack_depth,
                    stack_size=self.stack_size,
                    max_frame_depth=self.max_frame_depth
                )
            
            accumulate_ptx_stub = \
                stack_ptx_compiler.compile(
                    registry=registry,
                    instructions=accumulate_instructions, 
                    requests=accumulate_requests,
                    execution_limit=self.execution_limit,
                    max_ast_size=self.max_ast_size,
                    max_ast_to_visit_stack_depth=self.max_ast_to_visit_stack_depth,
                    stack_size=self.stack_size,
                    max_frame_depth=self.max_frame_depth
                )
            
            epilogue_ptx_stub = \
                stack_ptx_compiler.compile(
                    registry=registry,
                    instructions=epilogue_instructions, 
                    requests=epilogue_requests,
                    execution_limit=self.execution_limit,
                    max_ast_size=self.max_ast_size,
                    max_ast_to_visit_stack_depth=self.max_ast_to_visit_stack_depth,
                    stack_size=self.stack_size,
                    max_frame_depth=self.max_frame_depth
                )
            
            ptx_stubs = {
                'multiply':     multiply_ptx_stub,
                'accumulate':   accumulate_ptx_stub,
                'epilogue':     epilogue_ptx_stub,
            }

            rendered_ptx = inject.render_ptx(ptx_stubs)

            arch = get_compute_capability(device)
            module_cubin = Program(
                rendered_ptx,
                code_type="ptx", 
                options= \
                    ProgramOptions(
                        arch=f"sm_{arch}",
                        ptxas_options=['-O2', '-v'],
                    )
            ).compile(
                "cubin", 
                logs=sys.stdout,
                name_expressions=[function_name]
            )

            loaded_module = ObjectCode.from_cubin(module_cubin.code)
            symbol_map = {function_name: lowered_name}
            loaded_module._sym_map = symbol_map

            kernel = loaded_module.get_kernel(function_name)

            self.cubin_kernels_dict[dict_key] = kernel
        
        if debug:
            print(f'(Kermac Debug) Launching kernel: {function_name}')

        num_blocks_M = ceil_div(M, 128)
        num_blocks_N = ceil_div(N, 128)
        num_batches = L

        grid = (num_blocks_M, num_blocks_N, num_batches)
        config = LaunchConfig(grid=grid, block=256)

        ld_a = np.uint64(tensor_stats_a.leading_dimension_stride)
        batch_stride_a = np.uint64(tensor_stats_a.batch_stride)

        ld_b = np.uint64(tensor_stats_b.leading_dimension_stride)
        batch_stride_b = np.uint64(tensor_stats_b.batch_stride)

        ld_c = np.uint64(tensor_stats_c.leading_dimension_stride)
        batch_stride_c = np.uint64(tensor_stats_c.batch_stride)

        kernel_args = (
            M, N, K, L,
            a.data_ptr(),       ld_a,   batch_stride_a,
            b.data_ptr(),       ld_b,   batch_stride_b,
            out.data_ptr(),     ld_c,   batch_stride_c
        )

        if num_hypers:
            hyper_args = []
            for name, t in hyper_items:
                bs = np.uint64(0 if t.dim() == 0 or t.numel() == 1 else 1)  # scalar -> 0, [L] -> 0
                hyper_args.extend([t.data_ptr(), bs])
            kernel_args = (*kernel_args, *hyper_args)

        launch(stream, config, kernel, *kernel_args)

        return out



# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from platformdirs import user_cache_dir
import sys
from importlib.resources import files
from pathlib import Path
from importlib.metadata import version
import hashlib 
import uuid
import json
import tempfile
import cuda.pathfinder

def get_package_name():
    return "mm_kermac"

def get_package_version():
    try:
        pkg_version = version(get_package_name())
        return pkg_version
    except Exception as e:
        return f"Could not determine version: {e}"

def get_top_level_repo_dir(dir) -> Path:
    # directory *beside* the package (wheel layout)
    wheel_copy = files(get_package_name()).parent / dir
    # directory *beside* src/ (editable / repo checkout)
    repo_copy  = Path(__file__).resolve().parents[3] / dir

    for path in (repo_copy, wheel_copy):
        if path.is_dir():
            return path.resolve()

    raise FileNotFoundError("top level directory not found")

def get_include_local_cuda_dir() -> Path:
    return get_top_level_repo_dir('include')

def get_local_cuda_kernel_dir() -> Path:
    return get_top_level_repo_dir('kernels')

def get_include_paths():
    return [
        cuda.pathfinder.find_nvidia_header_directory('cccl'),       # for cstdint, etc..
        cuda.pathfinder.find_nvidia_header_directory('nvrtc'),      # for cuda_runtime_api.
        cuda.pathfinder.find_nvidia_header_directory('cudart'),
        get_include_local_cuda_dir(),                               # for common includes
        get_top_level_repo_dir('thirdparty/cutlass/include'),   # for cutlass includes
    ]

def cache_root() -> Path:
    """
    <user_cache_dir>/<your-package>/<env-id>/
    Guaranteed unique per virtual-env *and* persistent across runs.
    """
    try:
        # 1. A stable ID for THIS Python installation / venv
        env_id = hashlib.sha256(sys.prefix.encode()).hexdigest()[:12]

        # 2. RFC 6685 cache location that respects XDG / Windows / macOS rules
        base_dir = Path(user_cache_dir(f'{get_package_name()}')) / f'{get_package_version()}'
    
        # 3. Create sub-dir and a sentinel file on first use
        target = base_dir / env_id
        target.mkdir(parents=True, exist_ok=True)

        sentinel = target / "instance.json"
        if not sentinel.exists():
            sentinel.write_text(json.dumps({"uuid": str(uuid.uuid4())}))
        return target
    except Exception:               # Fallback: tmpdir
        return Path(tempfile.mkdtemp(prefix=f'{get_package_name()}-cache-'))


# SPDX-FileCopyrightText: 2025 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import threading
from typing import Any
import sys
import torch
import sqlite3
import os
import hashlib

from cuda.core.experimental import Device, Program, ProgramOptions

import mm_ptx.ptx_inject

from .function_db_key import *
from .function_db_value import *

from .paths import *

from .ptx_inject_types import DataTypeInfo

def get_compute_capability(device) -> str:
    if isinstance(device, torch.device):
        pt_device_id = device.index
        device = Device(pt_device_id)
    
    arch = "".join(f"{i}" for i in device.compute_capability)
    return arch

def hash_cuda_include_files(directories):
    # Initialize SHA-256 hash object
    hasher = hashlib.sha256()
    
    for directory in directories:
        # Walk through the directory
        for root, _, files in os.walk(directory):
            # Sort files for consistent hash across runs
            for file_name in sorted(files):
                # Check if file is a text file (e.g., ends with .cuh)
                if file_name.endswith('.cuh'):
                    file_path = os.path.join(root, file_name)
                    try:
                        # Read file in binary mode
                        with open(file_path, 'rb') as f:
                            # Update hash with file contents
                            while chunk := f.read(8192): # Read in 8KB chunks
                                hasher.update(chunk)
                    except (IOError, PermissionError) as e:
                        print(f"Error reading {file_path}: {e}")
    
    return hasher.hexdigest()


def ptx_db_create(db_path: str):
    """Create the SQLite database and table if it doesn't exist."""
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute('''
        CREATE TABLE IF NOT EXISTS function_db (
            key BLOB PRIMARY KEY,
            value BLOB
        )
    ''')
    conn.commit()
    conn.close()

def ptx_db_insert_entry(db_path: str, key: FunctionDBKey, value: FunctionDBValue):
    """Insert or replace an entry in the database using the key and value objects."""
    key_bytes = key.to_bytes()
    value_bytes = value.to_bytes()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute('INSERT OR REPLACE INTO function_db (key, value) VALUES (?, ?)', (key_bytes, value_bytes))
    conn.commit()
    conn.close()

def ptx_db_get_entry(db_path: str, key: FunctionDBKey) -> FunctionDBValue | None:
    """Retrieve a value from the database using the key object."""
    key_bytes = key.to_bytes()
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute('SELECT value FROM function_db WHERE key = ?', (key_bytes,))
    result = cursor.fetchone()
    conn.close()
    if result:
        return FunctionDBValue.from_bytes(result[0])
    return None

def compile_functions(
    arch,
    cuda_code,
    function_names,
    debug = False
):
    processed_cuda, num_injects = mm_ptx.ptx_inject.process_cuda(DataTypeInfo, cuda_code)
    annotated_ptx = Program(
        processed_cuda,
        code_type="c++", 
        options= \
            ProgramOptions(
                std="c++17",
                arch=f"sm_{arch}",
                device_as_default_execution_space=True,
                # diag_suppress cutlass: 64-D: declaration does not declare anything
                # diag_suppress cutlass: 1055-D: declaration does not declare anything
                diag_suppress=[64,1055],
                ptxas_options=['-v'] if debug else None,
                # Compiling these is a mess.
                define_macro=[
                    "__CUDA_NO_FP8_CONVERSIONS__", 
                    "__CUDA_NO_FP6_CONVERSIONS__",
                    "__CUDA_NO_FP4_CONVERSIONS__"
                ],
                include_path= get_include_paths()
            )
    ).compile(
        "ptx", 
        logs=sys.stdout,
        name_expressions=function_names
    )
    return annotated_ptx

class Singleton(type):
    """Metaclass for creating singleton classes."""
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]
    
class PtxInjectCache(metaclass=Singleton):
    """Singleton class mapping device IDs to lazily loaded modules/functions."""
    
    def __init__(self, debug = False):
        # A loaded kernel function is stored in device memory also
        # (device_id, function_name) -> Kernel
        self._lock = threading.Lock()
        if debug:
            print(f'(Kermac Debug) Using database at: {cache_root().resolve()}')
        include_dir = get_include_local_cuda_dir()
        kernels_dir = get_local_cuda_kernel_dir()
        self._hash_result = hash_cuda_include_files([include_dir, kernels_dir])
        if debug:
            print(f"(Kermac Debug) Combined hash of cuda source files: {self._hash_result}")
        self._cuda_version = str(torch.version.cuda)
        self._db_path = cache_root().resolve() / 'cache.db'
        os.makedirs(cache_root().resolve(), exist_ok=True)
        ptx_db_create(self._db_path)

    def get_function(
            self,
            device: Device, 
            function_name : str, 
            get_cuda_source_lambda,
            debug = False
        ) -> Any:
        device_id = device.device_id
        if device.compute_capability.major < 8:
            raise ValueError(f"Invalid device compute capability, (device:{device.compute_capability}, requrires at least:8.0")

        function_dict_key = (device_id, function_name)
        with self._lock:
            arch = get_compute_capability(device)
            db_key = FunctionDBKey(
                package_name=get_package_name(),
                package_version=get_package_version(),
                cuda_version=self._cuda_version,
                arch=arch,
                function_name=function_name,
                cuda_source_hash=self._hash_result
            )
            if debug: 
                print(f'(Kermac Debug) Checking database for ptx of function {function_name})')
            db_value = ptx_db_get_entry(self._db_path, db_key)
            if db_value is None:
                cuda_source = get_cuda_source_lambda()
                if debug: 
                    print(f'(Kermac Debug) Mapping does not exist for {db_key}')
                ptx = compile_functions(
                    arch,
                    cuda_source,
                    [function_name],
                    debug
                )
                lowered_name = ptx._sym_map[function_name]
                ptx_data = ptx.code
                db_value = FunctionDBValue(lowered_name, ptx_data)
                ptx_db_insert_entry(self._db_path, db_key, db_value)
                if debug: 
                    print(f'(Kermac Debug) Compiled entry and storing with key: {db_key}')
            else:
                if debug: 
                    print(f'(Kermac Debug) Mapping does exist for {db_key}')
                ptx_data = db_value.ptx_data
                lowered_name = db_value.lowered_name

            return ptx_data.decode('utf-8'), lowered_name
        